#include "T_op.h"
#include "../common/etvoTools.h"
#include <string>
#include "../common/etvoException.h"
#include<sstream>

using namespace std;

namespace etvo
{
  //static functions
  T_op T_op::E()
  {
    T_op e;
    e.setFromRw(Fmaxp::E());
    return e;
  }

  T_op T_op::Vee(unsigned v)
  {
    return dDd(0,v,1,0);
  }

  T_op T_op::Wedge(unsigned w)
  {
    return dDd(0,1,w,0);
  }
  T_op T_op::Delta(unsigned vw)
  {
    return dDd(0, vw, 0);
  }

  T_op T_op::Delta(unsigned v,unsigned w)
  {
    return dDd(0,v,w,0);
  }
 
  T_op T_op::deltaVar(const std::vector<int> & delays)
  {    
    for (int i = 0; i < delays.size() - 1; i++)
    {
      if (delays[i + 1] < (delays[i] - 1))
      {
        throw etvoException(12, "Delays are not correct");
      }
    }

    if (delays[0]<(delays[delays.size()-1] - 1))
    {
      throw etvoException(12, "Delays are not correct");
    }

    std::vector<int> Rw(delays.size());
    for (int i = 0; i < delays.size() ; i++)
    {
      Rw[i] = delays[i] + i;
    }
    int s = (int)delays.size();
    Fmaxp F(s,s,Rw);
    T_op t;
    t.setFromRw(F);
    return t;    
  }

 T_op T_op::delta(int t)
  {
    return dDd(t);
  }

  /// Neutral T_op
  T_op::T_op()
  {
    _fper = Fmaxp::E();
  }


  T_op::T_op(const dDd & term)
  {
    _fper = term.getRw();

  }

  void T_op::add(const T_op & op)
  {
    _fper = _fper.max(op._fper);    
  }

  void T_op::add(const dDd & term)
  {
     _fper = _fper.max(term.getRw());
  }

  std::pair<unsigned, unsigned> T_op::getPeriodicity() const
  {
    return _fper.getPeriodicity();
  }
  
  std::vector<dDd> T_op::getTerms() const
  {
    unsigned  v = getV();
	unsigned w = getW();
    vector<dDd> terms;

    for (int ti = 0; ti < (int)w; ti++)
    {
      if (_fper(ti) > _fper(ti - 1))
      {
        terms.push_back(dDd(_fper(ti), v,0,w, -w + 1 - ti));
      }
    }
    return terms;
  }
  unsigned T_op::getV() const
  {
    return (unsigned)_fper.getCodomPer();
  }

  unsigned T_op::getW() const
  {
	  return (unsigned)_fper.getDomPer();
  }
  
  T_op T_op::extendBy(unsigned mul) const
  {
      bool autoR = Fper::getAutoReductionState();
      Fper::setAutoReduction(false);
      Fper f=_fper.extendBy(mul);
      T_op result;
      result.setFromRw(f);
      Fper::setAutoReduction(autoR);
      return result;
  }

  void T_op::reduce()
  {
    _fper.reduce();

  }
  
  std::string T_op::toString() const
  {
    string s;
    std::vector<dDd> terms = this->getTerms();
    for (unsigned i = 0; i < terms.size(); i++)
    {
      s += terms[i].toString();
      if (i != terms.size() - 1) s += "+";
    }
    return s;
  }
  
  std::string T_op::toStringAsDeltaVar() const
  {
	 if (getPeriodicity().first != getPeriodicity().second)
	 {
		 return "(no desc.)";
	 }
    std::stringstream ss;
    Fmaxp rw = this->getRw();
    std::vector<int> d_shift(rw.getDomPer());
    for (int i = 0; i < rw.getDomPer(); i++)
    {
      d_shift[i] = rw(i) - i;
    }   
    ss << "d<" << d_shift[0];
    for (int i = 1; i < d_shift.size(); i++)
    {
      ss << "," << d_shift[i];
    }
    ss << ">";
    return ss.str();
  }

  int T_op::Rw(int ki) const
  {
    return _fper.getValue(ki);    
  }
  
  Fmaxp T_op::getRw() const
  {
    return _fper;
  }

  void T_op::setFromRw(const Fmaxp & f)
  {
    _fper = f;

  } 
     
  T_op T_op::operator+(const T_op & w) const
  {
    T_op resultat;
    resultat.setFromRw(_fper + w._fper);
    return resultat;
  }

  T_op T_op::oplus(const T_op & w) const
  {
    return (*this) + w;
  }
  

  T_op T_op::operator*(const T_op & w) const
  {
    T_op resultat;
    resultat.setFromRw(_fper*w._fper);
    return resultat;
  }
  
  T_op T_op::otimes(const T_op & w) const
  {
    return (*this)*w;
  }
  
  T_op T_op::inf(const T_op & w) const
  {
    T_op resultat;
    resultat.setFromRw(_fper.inf(w._fper));
    return resultat;
  }

  T_op T_op::lfrac(const T_op & w) const
  {
    T_op resultat;
    resultat.setFromRw(_fper.lfrac(w._fper));
    return resultat;
  }

  T_op T_op::rfrac(const T_op & w) const
  {
    T_op resultat;
    resultat.setFromRw(_fper.rfrac(w._fper));
    return resultat;
  }

bool T_op::operator==(const T_op & w) const
  {
    return _fper == w._fper;
  }

  bool T_op::operator!=(const T_op & w) const
  {
    return _fper != w._fper;
  }

  bool T_op::operator<=(const T_op & w) const
  {
    return _fper <= w._fper;
  }

  // Added 8/10/18 
  bool T_op::operator<(const T_op & w)const
  {
    return _fper < w._fper;
  }

  // Modification 8/10/18 -> uses <=
  bool T_op::operator>=(const T_op & w) const
  {
    return w<=(*this);
  }

  // Modification 8/10/18 -> uses <
  bool T_op::operator>(const T_op & w) const
  {
    return w < (*this);
  }  

  // 9/6/21
  T_op T_op::leastCausal(unsigned v, unsigned w)
  {
	  std::vector<int> seq(w);
	  for (unsigned i = 0; i < w; i++) seq[i] = (int)ceil((double)(v*i) / w);
	  Fmaxp Rw(w, v, seq);
	  T_op res;
	  res.setFromRw(Rw);
	  return res;
  }

  std::ostream & operator<<(std::ostream & f, const T_op & op)
  {
    f << op.toString();
    return f;
  }
}
