#ifndef DDD_H
#define DDD_H

#include<string>
#include<utility>
#include <iostream>

#include "../Fper/Fmaxp.h"

/*!
* \file  dDd.h
*
*/

/*! \namespace etvo
* namespace for ETVO classes
*/
namespace etvo
{
  class T_op; //anticipated declaration

/*!
* \class etvo::dDd
* \brief Class to describe terms in T[[g]] written d^tl.Delta_(m|b).d^tr= d^tl.V_m.d^tc.W_b.d^tr
*
* Formal series in T[[g]] can be seen as infinite sums of terms ( d^ti.Delta_(m|b).d^ti').g^n_i
*
* We can also decompose a dDd term into a central form (d^tli.V_m.d^tci.W_b.d^tri) where -m<tli<=0 and -b<tri<=0
*
* \author BC LH JT LARIS
* \version 2.0
*/
/// terms like d^tl.V_m.d^tc.W_b.d^tr

class dDd  
{
    public:
		/*! Class method to choose the canonical form of dDd terms
		*
		* \param val=0-left form 1-central form 2- right form
		*/
        /// set the type of canonical form 0=left 1=center 2=right
        static void setCanonForm(unsigned val=0);
		static unsigned getCanonForm();

		/// Create term d^tl.V_v.d^tc.W_w.d^tr
		dDd(int tl, unsigned int v, int tc,unsigned int w,int tr);
		/// Create term d^tl.V_v.d^0.W_w.d^tr
		dDd(int tl, unsigned int v, unsigned int w, int tr);
         /// Create term d^tl.D_(vw).d^tr=d^tl.V_m.d^0.W_m.d^tr
        dDd(int tl,unsigned int vw,int tr);

        /// d^0.V_1.d^tc.W_1.d^0 = d^tc
        dDd(int tc=0);
		        
		/// getter : gives tl in d^tl V_v d^tc W_w d^tr
        int getTl() const;
		/// getter : gives v in d^tl V_v d^tc W_w d^tr
        unsigned int getV() const;
		/// getter : gives tc in d^tl V_v d^tc W_w d^tr
        int getTc() const;
		/// getter : gives w in d^tl V_v d^tc W_w d^tr
		unsigned int getW() const;
		/// getter : gives tr in d^tl V_v d^tc W_w d^tr
		int getTr() const;		

        ///comparison of terms (with the same Periodicity)
        bool operator<=(const dDd & m) const;
        bool operator>=(const dDd & m) const;
        bool operator==(const dDd & m) const;


		 /// set to the canonical form (depends on the choice made by setCanonForm)
        void canon();
        /// set to the Left form [-w<tr<=0 and tc=0]
        void canonL();
        /// set to the Central form [-w<tr<=0 and -v<tl<=0]
         void canonC();
        /// set Right form [-v<tl<=0 and tc=0]
        void canonR();

        /// value of Release function Rw(ti) = ceil(((tr+ti)/W)+tc)*V+tl
        int Rw(int ki) const;
        /// returns function Rw
        Fmaxp getRw() const;

         /// Extension of d^tl V_v d^tc W_w d^tr
        T_op extendBy(unsigned mul) const;

        ///  periodicity as a pair <_b,_m>
        std::pair<unsigned,unsigned> getPeriodicity() const;
      
	    /// gives a string description of a dDd term
        /// \param nVer (default=0)   
        std::string toString(unsigned nVer=0) const;

		bool isEpsilon() { return false; }
    protected:
        /// set canonical form of dDd (default left form)
        static unsigned _canon;
        /// tl,v,tc,w,tr
        int _tl;
        unsigned int _v;
        int _tc;
		unsigned int _w;
        int _tr;
};
  /*! operator to print an element into the standard ostream */
    std::ostream & operator<<(std::ostream & f, const dDd & m);

} /// namespace ETVO
#endif // DDD_H
