#ifndef SERIESET_H
#define SERIESET_H
#include "../wrapperMMGD/seriesWrapper.h"
#include "../wrapperMMGD/matrixWrapper.h"
#include "../seriesEd/seriesEd.h"
#include "../seriesTg/seriesTg.h"
#include<iostream>
#include "../common/ISterm.h"


/*!
* \file  seriesET.h
*/

namespace etvo
{
	/*!
 * \class etvo::seriesET
 *
 * \brief Class for ultimately-periodic series in the semiring ET. 
 *
 * \author BC LH JT LARIS
 * \version 2.7
 */
  class seriesET:public ISterm
  {
  private:
    bool _QEdForm;  // true if Q in E[[d]] //false if Q in T[[g]]
	matrix<dDd> _Led;     
	matrix<seriesEd> _Qed;  // if _QEdForm -> Led.Qed.Red
	matrix<dDd> _Red;	
	matrix<gNg> _Ltg;     
	matrix<seriesTg> _Qtg;  // if !_QEdForm -> Ltg.Qtg.Rtg
	matrix<gNg> _Rtg;      	
	// internal usefull functions    	
	seriesET(const matrix<dDd> & Led, const matrix<seriesEd> Qed, const matrix<dDd> & Red);
	seriesET(const matrix<gNg> & Ltg, const matrix<seriesTg> Qtg, const matrix<gNg> & Rtg);

	void toQtg_polynomial_case();
	void toQed_polynomial_case();
	void toQtg_series_case();
	void toQed_series_case();

	void extendQed(unsigned mul);
	void extendQtg(unsigned mul);

	static seriesTg truncTg(const seriesTg &, int n, int gamma);
	static seriesEd truncEd(const seriesEd &, int n, int delta);
			
  public:
	  
	static seriesTg scale(const seriesTg & s, unsigned mul);
	static seriesEd scale(const seriesEd & s, unsigned mul);

    /// The epsilon description of seriesET
    static seriesET Epsilon();
	/// The Top description of seriesET
    static seriesET Top();
	/// The description of g0.d0 as seriesET
    static seriesET E();        
    /// Default initialization : epsilon  
    seriesET();
    /// Initialization as Top (true) OR E(false) 
	seriesET(bool TopNotE);
	/// Initialization of a seriesET from seriesEd 
	seriesET(const seriesEd & s);
	/// Initialization of a seriesET from seriesTg 
	seriesET(const seriesTg & s);
	   
	/// getter returning Led
	matrix<dDd> getLed() const;
	/// getter returning Red
	matrix<dDd> getRed() const;
	/// getter returning Qed 
	matrix<seriesEd> getQed() const;

	/// getter returning Ltg
	matrix<gNg> getLtg() const;
	/// getter returning Rtg
	matrix<gNg> getRtg() const;
	/// getter returning Qtg 
	matrix<seriesTg> getQtg() const;
	
    /// check if the current series is in Qed form (Q in E[[d]])
	bool isCoreEd() const { return _QEdForm; }
	/// check if the current series is in Qtg form (Q in T[[g]])
	bool isCoreTg() const { return !_QEdForm; }
	
	/// returns the Least Common multiple of gains in the terms of the current series
	void getLcmGainEd(unsigned int & mu, unsigned int& beta) const;
	void getLcmGainTg(unsigned int & v, unsigned int& w) const;
	void getLcmQSlope(int & g, int & d) const;

	/// to a form where the Core is in E[[d]]
	void toCoreEd();
	/// to a form where the Core is in T[[g]]	
	void toCoreTg();
	// extends the Core representation 
	void extendBy(unsigned mul);
	/// oplus sum of seriesET
	seriesET oplus(const seriesET & s) const;
	/// otimes product of seriesET
	seriesET otimes(const seriesET & s) const;
	/// Kleene star of seriesET
	seriesET star() const;
	/// infimum of seriesET
	seriesET inf(const seriesET & s) const;
	/// left residual
	seriesET lfrac(const seriesET & s) const;
	/// right residual
	seriesET rfrac(const seriesET & s) const;	

	/// sum of series in ET  : s1+s2
	seriesET operator+(const seriesET & s) const;
	/// product of series in ET  : s1*s2
	seriesET operator*(const seriesET  & s) const;
		
	/// CoreEd <-> CoreTg 
	void changeForm();

	/// check if it is a polynomial
	bool isPolynomial() const;
		
	/// check if it is a neutral seriesET
    bool isE() const;    

    /// leads to the canonical form 
    void canon();
    /// to Right form (all core series in right form)
    void toRight();
    /// to Left form (all core series in left form)
    void toLeft();
	
	void toLVectorEd(matrix<seriesEd> & LV, matrix<dDd> & R) const;
	void toRVectorEd(matrix<seriesEd> & RV, matrix<dDd> & L) const;

	void toLVectorTg(matrix<seriesTg> & LV, matrix<gNg> & R) const;
	void toRVectorTg(matrix<seriesTg> & RV, matrix<gNg> & L)const ;


	static etvo::matrix<seriesEd>  getMatNed(unsigned size);
	static etvo::matrix<seriesTg>  getMatNtg(unsigned size);

	static seriesET toCausal(const seriesET & s);

	/// returns the string description of a series in ET. 
    std::string toString(unsigned f=0) const;	
	

	friend seriesET etg(int n, bool QedForm );
	friend seriesET etd(int t, bool QedForm );	
	friend seriesET etm(unsigned m, bool QedForm );	
	friend seriesET etb(unsigned b, bool QedForm );	
	friend seriesET etm(const std::vector<unsigned> & seq);
	friend seriesET etb(const std::vector<unsigned> & seq);
	friend seriesET etd(const std::vector<int> & seqDelays);
	friend seriesET etv(unsigned v, bool QedForm );
	friend seriesET etw(unsigned w, bool QedForm);

  };
  
  std::ostream & operator<<(std::ostream & f, const seriesET & s);


  // necessary for matrix<T> computations.
  seriesET star(const seriesET & s);
  seriesET oplus(const seriesET &, const seriesET &);
  seriesET inf(const seriesET &, const seriesET &);
  seriesET otimes(const seriesET &, const seriesET &);
  seriesET lfrac(const seriesET &, const seriesET &);
  seriesET rfrac(const seriesET &, const seriesET &);

  /// element g^n in ET
  seriesET etg(int n,bool QedForm=true);
  /// element d^t in ET
  seriesET etd(int t,bool QedForm = true);
  /// element mu_m in ET (Event multiplier)
  seriesET etm(unsigned m,bool QedForm = true);
  seriesET etm(const std::vector<unsigned> & seq);
  seriesET etb(const std::vector<unsigned> & seq);
  seriesET etd(const std::vector<int> & seqDelays);

  /// element beta_b in ET (EVent divider)
  seriesET etb(unsigned b, bool QedForm = true);
  /// element vee_v in ET (Time multiplier)
  seriesET etv(unsigned v, bool QedForm = true);
  /// element wedge_w in ET (Time divider)
  seriesET etw(unsigned w, bool QedForm = true);

 
  /*
  /// element mu var mu_<seq> in ET
  seriesET etm(const std::vector<unsigned> & seq);
  /// element beta var beta_<seq> in ET
  seriesET etb(const std::vector<unsigned> & seq);
  
  // necessary for matrix<T> computations.
  seriesET star(const seriesET & s);
  seriesET oplus(const seriesET &, const seriesET &);
  seriesET inf(const seriesET &, const seriesET &);
  seriesET otimes(const seriesET &, const seriesET &);
  seriesET lfrac(const seriesET &, const seriesET &);
  seriesET rfrac(const seriesET &, const seriesET &);

  */
} // end namespace etvo

#endif // SERIESET_H
