#ifndef POLYTG_H
#define POLYTG_H
#include "Tg.h"
#include "../common/ISterm.h"

#include "../wrapperMMGD/polyWrapper.h"
#include "../wrapperMMGD/matrixWrapper.h"

namespace etvo
{
 class seriesTg;
 /*!
 * \class etvo::polyTg
 *
 * \brief Class for polynomials in the semiring T[[g]].
 *
 * epsilon and top element exist.
 *
 * \author BC LH JT LARIS
 * \version 2.0
 */

class polyTg:public ISterm
{
public:
  /// Epsilon element
  static polyTg Epsilon();
  /// Top element
  static polyTg Top();
  /// neutral element
  static polyTg E();

  /// initialised with Epsilon element
  polyTg();
  ///  specific initialization : polyTg(true) set to Top, polyTg(false) set to E 
  polyTg(bool TopNotE); //true Top false E
  ///  initialised with one monomial m
  polyTg(const Tg & m);
  /// initialization to a polynomial created from a collection of Tg terms w.g^n. 
  /// The initialization leads to a canonical form where Tg terms are sorted 
  polyTg(const std::vector<Tg> & v);

  /// set to the canonical form
  void canon();
  /// Check if a polyTg is in canonical form
  bool isCanon() const;
  /// check if is equal to E()=g0.d0
  bool isE() const;
  /// Sum of polynomials in T[[g]]   
  polyTg operator+(const polyTg &p) const;
  /// Sum of polynomials in T[[g]]   
  polyTg oplus(const polyTg &p) const;
  /// Sum of polynomials in T[[g]] via a Core Decomposition (see J.Trunk Thesis)
  polyTg oplusCD(const polyTg &p) const;
  /// Sum of a polynomial in T[[g]] with a monomial in  T[[g]]
  polyTg operator+(const Tg & m)const;
  /// The current polynomial pcur is modified as pcur=pcur+m
  void add(const Tg & m);
  /// Lies on polyTg::add(const Tg & m) method
  polyTg operator+=(const Tg & m);

  /// Product of polynomials in T[[g]]
  polyTg operator*(const polyTg & p) const;
  /// Product of polynomials in T[[g]]
  polyTg otimes(const polyTg &p) const;
  /// Product of one polynomial by one monomial in T[[g]]
  polyTg operator*(const Tg & m) const;
  static polyTg otimes(const Tg & m, const polyTg & p);
  /// Product of polynomials in T[[g]] via a Core Decomposition (see j.Trunk thesis)
  polyTg otimesCD(const polyTg &p) const;
  
  /// Infimum of two polynomials in T[[g]]
  polyTg inf(const polyTg & p) const;
  /// Infimum of two polynomials in T[[g]] via a Core Decomposition (see J.Trunk thesis)
  polyTg infCD(const polyTg & p) const;
  
  /// Kleene star of a polynomial in T[[g]], the result is a series in T[[g]]
  seriesTg star() const;
	
  /// Computation of the left-multiplication residuation: 
  polyTg lfrac(const polyTg & p) const;
  /// Computation of the left-multiplication residuation via a Core Decomposition 
  polyTg lfracCD(const polyTg & p) const;
  /// Computation of the left-multiplication residuation: 
  polyTg lfrac(const Tg & m) const;
  /// Computation of the right-multiplication residuation:
  polyTg rfrac(const polyTg & p) const;
  /// Computation of the right-multiplication residuation via a Core Decomposition 
  polyTg rfracCD(const polyTg & p) const;
  /// Computation of the right-multiplication residuation:	
  polyTg rfrac(const Tg & m) const;
  /// Check equality
  bool operator==(const polyTg &)const;
  /// Check difference
  bool operator!=(const polyTg &)const;
  /// Check order on polynomials in E[[d]]
  bool operator<=(const polyTg &) const;
  /// Check order on polynomials in E[[d]]
  bool operator>=(const polyTg &) const;

  /// Creates a polynomial in T[[g]] from a polynomial in MinMax[[g,d]]
  static polyTg toPolyTg(const poly & p);
  /// The zero-slice polynomial in MinMax[[g,d]]
  poly toPoly() const;
  /// returns the projection of p into the set of causal series in T[[g]] (not reliable yet)
  static polyTg toCausal(const polyTg & p);
  /// When the current polynomial is different from p, gives the first different monomial
  /// Otherwise, if both polynomials are equals, returns Tg::E()=g0.d0
  Tg getFirstDif(const polyTg & p) const;  
  /// Do not use it. Use polyTg::star(). Only for DEBUGGING purpose.
  polyTg transientStar(int Tmax) const;
  /// Gives the maximal gain
  void getMaxGain(unsigned int & vee, unsigned int & wedge) const;
  /// Gives th Least Common multiple of gains
  void getLcmGain(unsigned int & vee, unsigned int & wedge) const;
  /// Returns the periodicity as a pair
  std::pair<unsigned int, unsigned int> getPeriodicity() const;
  /// return the monomials as a collection of Tg terms
  std::vector<Tg> getTerms() const;
  /// remove term number i in the polynomial
  void removeTerm(unsigned idx);
  /// Returns a copy of monomial in position idx in the polynomial
  Tg operator[](unsigned idx) const;
  /// Returns the size = the number of monomials. For Epsilon and Top, size=0.
  unsigned int size() const;
  /// returns a string that gives the description of the current polynomial. Is depending on the canonical form of dDd terms
  std::string toString() const;
  /// returns a description of the current polynomial (the gain must be 1)
  std::string toStringAsDeltaVar() const;
   
  /// returns the Core matrix<poly> (in MinMax[[g,d]]) of the current polynomial
  matrix<poly> getCore(unsigned ratio = 1) const;
  /// computes the recomposition of a polyTg polynomial from a Core Decomposition core.
  static polyTg coreToPolyTg(const matrix<poly>& core);

  static etvo::matrix<poly> getMatN(unsigned size);
  /// returns the maximal Core matrix<poly> (in MinMax[[g,d]]) of the current polynomial
  matrix<poly> getCoreMax(unsigned ratio = 1) const;
 
  /// used in the creation of POV-Ray script for a polyTg object
  void toPov(graphicPR::PovRay & pov, graphicPR::PovRay::Color c);

private:
  void sort();
  void simplify();
	std::vector<Tg> _poly;
};
std::ostream & operator<<(std::ostream & st,const polyTg & p);
}/// end namespace

#endif // POLYTG_H
